/*
 * Galactic Bloodshed, copyright (c) 1989 by Robert P. Chansky, 
 * smq@ucscb.ucsc.edu, mods by people in GB_copyright.h.
 * Restrictions in GB_copyright.h.
 *
 *  moveship -- moves specified ship according to its orders.
 *	also deducts fuel from the ship's stores.
 */

#include "GB_copyright.h"
#define EXTERN extern
#include "vars.h"
#include "ships.h"
#include "races.h"
#include "doturn.h"
#include "buffers.h"
#include <math.h>


/* amount to move for each dir level. I arrived on these #'s only after
	hours of dilligent tweaking */
  /* amount to move for each directory level  */
float MoveConsts[] = { 500.0, 200.0, 30.0 };
  /* amnt to move for each ship speed level (ordered) */
float SpeedConsts[] = { 0.0, 0.61, 1.26, 1.50, 1.73, 1.81, 1.90, 1.93, 1.96, 1.97};
  /* amount of fuel it costs to move at speed level */


Moveship(shipno,s)
int shipno;
shiptype *s;
{
 double stardist,movedist,truedist,dist,xdest,ydest,sn,cs;
 float fuse,mfactor,heading;
 int destlevel,deststar=0,destpnum=0;
 boolean move_err = 0;
 shiptype *dsh;
 startype *ost,*dst;
 planettype *opl,*dpl;

if(fabs(s->xpos) > 999999999. || fabs(s->ypos) > 999999999.)
	{
	printf("we got a problem here!\n");
	return;
	}

if(s->hyper_drive.on) { 	/* do a hyperspace jump */
/* printf("hyperspace ship #%d\n", shipno); */
	if(s->hyper_drive.ready) {
/* uses up fuel to jump. This is dependent on the mass of the ship - if you don't have
	enough fuel for the jump - jump is cancelled and a message sent to the owner */
	 dist = sqrt( Distsq(s->xpos, s->ypos, Stars[s->deststar]->xpos, Stars[s->deststar]->ypos));
	fuse = HYPER_DRIVE_FUEL_USE * s->mass 
		* (dist/HYPER_DIST_FACTOR) * (dist/HYPER_DIST_FACTOR);
	if(s->fuel < fuse) {
		  sprintf(telegram_buf,
			   "%s #%d %s at system %s does not have %.1ff to do hyperspace jump.\n",
			   Shipnames[s->type], shipno, s->name, prin_ship_orbits(s), fuse);
			push_message(TELEG_PLAYER_AUTO, s->owner, telegram_buf);
		s->hyper_drive.on = 0;
		return;
		}

   		s->fuel -= fuse;
   		s->mass -= fuse * MASS_FUEL;

   		heading = atan2( Stars[s->deststar]->xpos - s->xpos, Stars[s->deststar]->ypos - s->ypos);
		sn = sin(heading);
		cs = cos(heading);
	 	s->xpos = Stars[s->deststar]->xpos + 	sn * 0.9 * SYSTEMSIZE;
	 	s->ypos = Stars[s->deststar]->ypos + cs * 0.9 * SYSTEMSIZE;

		s->whatdest = LEVEL_UNIV;
		s->whatorbits = LEVEL_STAR;
		s->storbits = s->deststar;

		s->hyper_drive.on = 0;
		s->hyper_drive.ready = 0;
		s->hyper_drive.charge = 0;
		  sprintf(telegram_buf,"%s #%d %sarrived at system %s.\n",
			   Shipnames[s->type], shipno, s->name,
				prin_ship_orbits(s));
			str_cat(telegram_buf, buf);
		push_message(TELEG_PLAYER_AUTO, s->owner, buf);
		} else {
			if(s->hyper_drive.charge==HYPER_DRIVE_READY_CHARGE)
				s->hyper_drive.ready = 1;
			else		
				s->hyper_drive.charge += 1;
			  }
	return;
	}

 if (s->speed && !s->is_docked && s->is_alive && (s->whatdest!=LEVEL_UNIV
		|| s->navigate.on) ) {

 /*printf("moving ship #%d\n",shipno);*/

    /* subtract fuel from the ship */
   fuse = 0.5 * s->speed * (1 + s->protect.evade) * s->mass * FUEL_USE;

   if (s->fuel < fuse) {
	  /* ship is out of fuel; do whatever it is to do */
	if (!s->notified) {
	  s->notified = 1;
	  s->fuel = 0.0;		/* in case it's negative */
	  msg_OOF(s, shipno, telegram_buf);	/* send OOF notify */

		/* kill the ship if we lost it and it's expendable */
	  if (s->whatorbits == LEVEL_UNIV && Shipdata[s->type][ABIL_COST]<=50) {
	    sprintf(telegram_buf,"Ship #%d has been lost in deep space.\n",shipno);
	push_message(TELEG_PLAYER_AUTO, s->owner, telegram_buf);
	    kill_ship(s);
	  }
	}
	return;
   }

/*******   move the ship towards dest ******/
	if(s->navigate.on){   /* follow navigational orders */
	heading = .0174329252*s->navigate.bearing;
   mfactor = SHIP_MOVE_SCALE * (1. - .01 * s->rad) * (1. - .01 * s->damage) 
        * SpeedConsts[s->speed] * MoveConsts[s->whatorbits];

   	s->fuel -= fuse;
   	s->mass -= fuse * MASS_FUEL;

	sn = sin(heading);
	cs = cos(heading);
   	xdest = sn * mfactor;
   	ydest = cs * mfactor;

   	s->xpos += xdest;
   	s->ypos -= ydest;

	s->navigate.turns --;
	if(s->navigate.turns == 0) s->navigate.on = 0; 

} else {  /*		navigate is off            */

   switch (s->whatdest) {
      case LEVEL_STAR:
	 destlevel = LEVEL_STAR;
	 deststar = s->deststar;
	 xdest = Stars[deststar]->xpos;
	 ydest = Stars[deststar]->ypos;
	 break;
      case LEVEL_PLAN:
	 destlevel = LEVEL_PLAN;
	 deststar = s->deststar;
	 destpnum = s->destpnum;
	 xdest = Stars[deststar]->xpos + planets[deststar][destpnum]->xpos;
	 ydest = Stars[deststar]->ypos + planets[deststar][destpnum]->ypos;
	 if(sqrt( Distsq(s->xpos, s->ypos, xdest, ydest)) <= DIST_TO_LAND)
	   destlevel = LEVEL_UNIV;
	 break;
      case LEVEL_SHIP:
	 dsh = ships[s->destshipno];
	 destlevel = LEVEL_SHIP;
	 deststar = dsh->storbits;
	 destpnum = dsh->pnumorbits;
	 xdest = dsh->xpos;
	 ydest = dsh->ypos;
	 if(sqrt( Distsq(s->xpos, s->ypos, xdest, ydest))==0.0)destlevel = LEVEL_UNIV;
	 break;
      default:
	 printf("wow, seriously bad error.\n");
	 return;
	 break;
   }
   dst = Stars[deststar];
   ost = Stars[s->storbits];
   dpl = planets[deststar][destpnum];
   opl = planets[s->storbits][s->pnumorbits];


/*  update new xpos,ypos */
   heading = atan2( xdest-s->xpos, 
	((ydest - s->ypos)==0.0) ? 0.000001 : (ydest-s->ypos) );
   mfactor = SHIP_MOVE_SCALE * (1. - .01 * s->rad) * (1. - .01 * s->damage)
       * SpeedConsts[s->speed] * MoveConsts[s->whatorbits];


   truedist = movedist = sqrt( Distsq(s->xpos, s->ypos, xdest, ydest));


    /* keep from ending up in the middle of the system. */
	/* Change the distance so that we
	   don't go all the way into the system. */
   if (s->whatorbits==LEVEL_UNIV ||
	  (destlevel==LEVEL_STAR && s->storbits!=deststar) ||
	  (destlevel==LEVEL_PLAN && s->storbits!=deststar) &&
	   truedist >= SYSTEMSIZE)
   	movedist -= SYSTEMSIZE * 0.90;
   else if (destlevel==LEVEL_PLAN && s->whatorbits==LEVEL_STAR &&
		s->storbits==deststar && truedist >= PLORBITSIZE)
	 /* we're in a system, don't go all the way into a planetary
	    orbit */
	movedist -= PLORBITSIZE * 0.90;


   if (s->whatdest==LEVEL_SHIP && s->owner!=ships[s->destshipno]->owner
	&& !isset(races[s->owner-1]->allied, ships[s->destshipno]->owner)
   	   && truedist > SYSTEMSIZE*4.0) {
	  /* (an allied ship lets you follow it..) */
	  /* we lost sight of the destination ship. */
	  /* should change SYSTEMSIZE*1.0 to calculated gun range.. */
	s->whatdest = LEVEL_UNIV;
	s->protect.evade = 0;
   	sprintf(telegram_buf,"Telecomm from ship #%d at %s\n\n", 
				shipno, prin_ship_orbits(s));
    	sprintf(buf,"%s #%d %s lost sight of destination ship #%d.\n", 
			Shipnames[s->type],shipno,s->name,s->destshipno);
		str_cat(telegram_buf, buf); 
	push_message(TELEG_PLAYER_AUTO, s->owner, telegram_buf);
	return;
   }


   if (truedist > DIST_TO_LAND) {
   	s->fuel -= fuse;
   	s->mass -= fuse * MASS_FUEL;

		/* dont overshoot */
	sn = sin(heading);
	cs = cos(heading);
   	xdest = sn * mfactor;
   	ydest = cs * mfactor;
    	if (hypot(xdest, ydest) > movedist) {
		xdest = sn * movedist;
		ydest = cs * movedist;
	}
   	s->xpos += xdest;
   	s->ypos += ydest;
   }

/*****  check if far enough away from object it's orbiting to break orbit *****/

  if (s->whatorbits==LEVEL_PLAN) {
		dist = sqrt( Distsq(s->xpos, s->ypos, 
			ost->xpos+opl->xpos, ost->ypos+opl->ypos ) );
/*	 printf("dist from orbts planet is %lf\n", dist);*/
	 if (dist > PLORBITSIZE) {
		s->whatorbits = LEVEL_STAR;
/*	 	printf("ship #%d departed from planet /%s/%s.\n",shipno, ost->name, ost->pnames[s->pnumorbits]);*/
	 }

   } else if (s->whatorbits==LEVEL_STAR) {
	dist = sqrt( Distsq(s->xpos, s->ypos, ost->xpos, ost->ypos ) );
/*	 printf("star orbits dist is %lf\n", dist); */
	 if (dist > SYSTEMSIZE) {
		s->whatorbits = LEVEL_UNIV;
		s->protect.evade = 0;
/*	 	printf("ship #%d departed from system %s.\n",shipno, ost->name); */
	 }
   }

	/*******   check for arriving at destination *******/

   if (destlevel==LEVEL_STAR || (destlevel==LEVEL_PLAN
	&& !(s->whatorbits!=LEVEL_UNIV && s->storbits==deststar) ) ) {
       		/* not in same star system as deststar */

	 stardist = sqrt(Distsq(s->xpos,s->ypos,dst->xpos,dst->ypos));
/*	 printf("dist from dest syst is %g\n",stardist); */
	 if (stardist <= SYSTEMSIZE * 1.5) {
	   if (s->type == STYPE_POD)
		s->notified = 1;	/* signal to explode */
	    /* mark as explored & inhabited by that player */
	   setbit(dst->explored, s->owner);
	   setbit(dst->inhabited, s->owner);
/*	   printf(" ship #%d arrived at system %s.\n",shipno,dst->name); */
	   s->whatorbits = LEVEL_STAR;
	   s->storbits = deststar;
	   if (destlevel == LEVEL_STAR) {
		  sprintf(telegram_buf,"%s #%d %sarrived at system %s.\n",
			   Shipnames[s->type], shipno, s->name,
			prin_ship_orbits(s));
			str_cat(telegram_buf, buf);
	push_message(TELEG_PLAYER_AUTO, s->owner, buf);
		/* clear orders if the ship is not headed to a planet in
		   this system */
	   	  s->whatdest = LEVEL_UNIV;
	   }
	 }

   } else if ( destlevel==LEVEL_PLAN && 
		 (s->whatorbits!=LEVEL_UNIV && deststar==s->storbits  &&
		!(s->whatorbits==LEVEL_PLAN && s->pnumorbits==destpnum)) ) {
	/* headed for a planet in the same system, & not already there.. */
		dist = sqrt( Distsq(s->xpos, s->ypos, 
			dst->xpos+dpl->xpos, dst->ypos+dpl->ypos ) ) ;
/*	 printf("dist from target planet is %lf\n",dist ); */
	 if (dist<=PLORBITSIZE) {
		 /* mark planet as explored by that player */
	   dpl->info[s->owner-1].explored = 1;
		 /* mark sun as explored too, if the ship might have bypassed it
		    (we are trying to be thorough here) */
	   setbit(dst->explored, s->owner);
	   setbit(dst->inhabited, s->owner);
	   s->whatorbits = LEVEL_PLAN;
	   s->pnumorbits = destpnum;	

	   sprintf(telegram_buf,"Telecomm from ship #%d\n\n",shipno);
/* 	   printf("%s #%d arrived at planet /%s/%s.\n", Shipnames[s->type],
		shipno, ost->name, ost->pnames[s->pnumorbits]); */

	if(dist<=DIST_TO_LAND) {
	   sprintf(buf,"%s #%d %sarrived at planet %s.\n This ship is close enough to land.\n",Shipnames[s->type],shipno, s->name, prin_ship_orbits(s));
	} else {
	   sprintf(buf,"%s #%d %sarriving at planet %s.\n This ship is %4.1f units away.\n",
				Shipnames[s->type],shipno, s->name, prin_ship_orbits(s),dist);
		}


	str_cat(telegram_buf, buf);
	   if (s->type==STYPE_ASS) {
	   	sprintf(buf,"\nEnslavement of the planet is now possible.\n");
		str_cat(telegram_buf, buf);
			}
	push_message(TELEG_PLAYER_AUTO, s->owner, telegram_buf);
	 }

  } 

 }

} /* if 'destination' orders */

}



/* 
 * insert the ship into wherever it's supposed to be.
 *  this is done every turn, for every ship; as a bonus it puts them in
 *  alphabetical order.
 */
insert_sh(sdata, star, pl, s, shipno)
struct stardata *sdata;
startype *star;
planettype *pl;
shiptype *s;
int shipno;
{

 if (s->is_alive)
    switch (s->whatorbits) {
	case LEVEL_UNIV:
		s->nextship = sdata->ships;
		sdata->ships = shipno;
		break;
	case LEVEL_STAR:
		s->nextship = star->ships;
		star->ships = shipno;
		break;
	case LEVEL_PLAN:
		s->nextship = pl->ships;
		pl->ships = shipno;
		break;
	default:
		break;
    }

}


/* deliver an "out of fuel" message.  Used by a number of ship-updating
 *  code segments; so that code isn't duplicated.
 */
msg_OOF(s, shipno, tbuf)
shiptype *s;
int shipno;
char *tbuf;
{
 char buf[200];

	sprintf(buf,"%s #%d is out of fuel at %s\n", Shipnames[s->type],
				shipno, prin_ship_orbits(s));
	push_message(TELEG_PLAYER_AUTO, s->owner, buf);
}


insert_sh_univ(sdata, s, shipno)
shiptype *s;
struct stardata *sdata;
{
	s->nextship = sdata->ships;
	sdata->ships = shipno;
}


insert_sh_star(star, s, shipno)
startype *star;
shiptype *s;
int shipno;
{
	s->nextship = star->ships;
	star->ships = shipno;
}


insert_sh_plan(pl, s, shipno)
planettype *pl;
shiptype *s;
int shipno;
{
	s->nextship = pl->ships;
	pl->ships = shipno;
}
